!+
! Program:  epframes/test/checkoutput
!
! Authors:  Michael Freyberg - XMM-SSC-MPE
!          
! Description: Tests output and functionality of epframes
! 
!- 

program checkoutput_OBT

use dal
use errorhandling
! use epframes_testio

implicit none

type(DataSetT)  :: inevents, ingti
type(TableT)    :: eventstable, expotable, gtitable
type(ColumnT)   :: timeCol, rawXCol, rawYCol, flagCol, ccdnrCol, &
                   startCol, stopCol, frameCol, phaCol, fracExpCol

real(kind=double),  dimension(:), pointer :: time, start, stop
real(kind=single),  dimension(:), pointer :: fracExp
integer(kind=Int8) ,dimension(:), pointer :: ccdnr
integer(kind=Int16),dimension(:), pointer :: rawX, rawY, pha
integer(kind=Int32),dimension(:), pointer :: flag, frame

real(kind=double)                 :: deltat, halfdt, maxerr
integer                           :: i, nevrows, ngtirows, nr
integer, parameter                :: nt = 25, ns = 6180
integer(kind=Int8)                :: CCDno
integer,parameter                 :: itest = 6  ! OBT
logical(kind=bool)                :: error_in_test, empty_frames
character(len=20),dimension(1:7)  :: list = (/"rawevents_no.dat   ", &
                                              "rawevents_IM.dat   ", &
                                              "rawevents_TI.dat   ", &
                                              "rawevents_BU.dat   ", &
                                              "rawevents_SW.dat   ", &
                                              "rawevents_OBT.dat  ", &
                                              "rawevents_OBT_N.dat"/)
character(len=10),dimension(1:7)  :: extlist = (/"EVENTS", &
                                                 "EVENTS", &
                                                 "EVENTS", &
                                                 "EVENTS", &
                                                 "EVENTS", &
                                                 "EVENTS", &
                                                 "EVENTS"/)
character(len=20),dimension(1:7)  :: gtilist = (/"gti_no.dat   ", &
                                                 "gti_IM.dat   ", &
                                                 "gti_TI.dat   ", &
                                                 "gti_BU.dat   ", &
                                                 "gti_SW.dat   ", &
                                                 "gti_OBT.dat  ", &
                                                 "gti_OBT_N.dat"/) 

character(len=20)                 :: expname, gtiname, readoutmode

real(kind=double)   :: time_N(1:nt+1), diff_ev_ex_Y, diff_ev_ex_N
! real(kind=single)   :: fracExp_N(1:nt+1)
integer(kind=Int8)  :: ccdnr_N(1:nt+1)
integer(kind=Int16) :: rawX_N(1:nt+1), rawY_N(1:nt+1), pha_N(1:nt+1)
integer(kind=Int32) :: flag_N(1:nt+1), frame_N(1:nt+1)




maxerr = 1.e-9 ! allowed discrepancy in time computation [1ns]
maxerr = 1.e-4 ! allowed discrepancy in time computation [0.1ms]

error_in_test = .false.
empty_frames  = .false.

  if (itest == 1 .or. itest == 2) readoutmode = 'IMAGING'
  if (itest == 3) readoutmode = 'TIMING'
  if (itest == 4) readoutmode = 'BURST'
  if (itest == 5) readoutmode = 'SmallWindow'
  if (itest == 6) readoutmode = 'FullFrame invalidOBT'

  write (*,*) 'testing epframes ',trim(readoutmode),' mode output:'
  
  !-----------------------------------------
  write (*,*) 'testing ',list(itest),' ...'
  !-----------------------------------------

  inevents = dataSet(list(itest+1), read)

  !-----------------------------------

  eventstable = table(inevents,extlist(itest+1))

  if (.not.hasColumn(eventstable,"TIME")  .or. &
      .not.hasColumn(eventstable,"FRAME") .or. &
      .not.hasColumn(eventstable,"RAWX")  .or. &
      .not.hasColumn(eventstable,"RAWY")  .or. &
      .not.hasColumn(eventstable,"PHA")   .or. &
      .not.hasColumn(eventstable,"CCDNR") .or. &
      .not.hasColumn(eventstable,"FLAG")) then
    error_in_test = .true. 
    call warning("noEvColumn", &
                 "Events extension: column absent or incorrectly named")
  endif

    TimeCol  = column(eventstable, "TIME" ,read)
    FrameCol = column(eventstable, "FRAME",read)
    rawXCol  = column(eventstable, "RAWX" ,read)
    rawYCol  = column(eventstable, "RAWY" ,read)
    phaCol   = column(eventstable, "PHA"  ,read)
    flagCol  = column(eventstable, "FLAG" ,read)
    ccdnrCol = column(eventstable, "CCDNR",read)

    nevrows = numberofRows(eventstable)
    nr = nevrows-1
    write(errstr,*) 'Event file version N: number of entries:',nevrows   
    call message(AppMsg,Sparse)

    time  => real64Data(timeCol)
    frame => int32Data(frameCol)
    rawX  => int16Data(rawXCol)
    rawY  => int16Data(rawYCol)
    pha   => int16Data(phaCol)
    flag  => int32Data(flagCol)
    ccdnr => int8Data(ccdnrCol)

    time_N(1:nt)   = time(ns:ns-1+nt)
    frame_N(1:nt)  = frame(ns:ns-1+nt)
    rawx_N(1:nt)   = rawx(ns:ns-1+nt)
    rawy_N(1:nt)   = rawy(ns:ns-1+nt)
    pha_N(1:nt)    = pha(ns:ns-1+nt)
    flag_N(1:nt)   = flag(ns:ns-1+nt)
    ccdnr_N(1:nt)  = ccdnr(ns:ns-1+nt)

    time_N(nt+1)   = time(nr)
    frame_N(nt+1)  = frame(nr)
    rawx_N(nt+1)   = rawx(nr)
    rawy_N(nt+1)   = rawy(nr)
    pha_N(nt+1)    = pha(nr)
    flag_N(nt+1)   = flag(nr)
    ccdnr_N(nt+1)  = ccdnr(nr)

    diff_ev_ex_N = time(nr)

    write(expname,'(a)') "EXPOSURE"
    expotable = table(inevents,expname)
    TimeCol  = column(expotable, "TIME" ,read)
    FracExpCol = column(expotable, "FRACEXP" ,read)
    time => real64Data(timeCol)
    fracExp => real32Data(fracExpCol)

    nevrows = numberofRows(expotable)
    nr = nevrows-1
    write(errstr,*) 'Last entry N:  TIME, FRACEXP:',TIME(nr),FRACEXP(nr)
    call message(AppMsg,Sparse)
    diff_ev_ex_N = TIME(nr)-diff_ev_ex_N

  call release(inevents)



  inevents = dataSet(list(itest), read)
  eventstable = table(inevents,extlist(itest))

  if (.not.hasColumn(eventstable,"TIME")  .or. &
      .not.hasColumn(eventstable,"FRAME") .or. &
      .not.hasColumn(eventstable,"RAWX")  .or. &
      .not.hasColumn(eventstable,"RAWY")  .or. &
      .not.hasColumn(eventstable,"PHA")   .or. &
      .not.hasColumn(eventstable,"CCDNR") .or. &
      .not.hasColumn(eventstable,"FLAG")) then
    error_in_test = .true. 
    call warning("noEvColumn", &
                 "Events extension: column absent or incorrectly named")
  endif





  write(expname,'(a)') "EXPOSURE"
  expotable = table(inevents,expname)

  if (.not.hasColumn(expotable,"TIME")    .or. &
!     .not.hasColumn(expotable,"TIMEDEL") .or. &
      .not.hasAttribute(expotable,"TIMEDEL") .or. &
      .not.hasColumn(expotable,"FRACEXP")) then
    error_in_test = .true. 
    call warning("noExpoColumn", &
               "Exposure extension: column absent or incorrectly named")   
  endif

  TimeCol  = column(expotable, "TIME" ,read)
  FracExpCol = column(expotable, "FRACEXP" ,read)
  time => real64Data(timeCol)
  fracExp => real32Data(fracExpCol)

  nevrows = numberofRows(expotable)
  nr = nevrows-1

  diff_ev_ex_Y = TIME(nr)

  if (nevrows < 1) then
    write(errstr,*) 'Test not successful for ITEST = ',itest, &
       ', empty input file'
    error_in_test = .true. 
    call warning("emptyEXPOFile")
  elseif (nevrows .ne. 184002) then
    error_in_test = .true.
    write(errstr,*) 'EXPOSURE does not have 184002 entries, but:',nevrows
    call warning("wrongEXPOFileSize")    
    write(errstr,*) 'First entry: TIME, FRACEXP:',TIME(0),FRACEXP(0)
    call message(AppMsg,Sparse)
    write(errstr,*) 'Last entry:  TIME, FRACEXP:',TIME(nr),FRACEXP(nr)
    call message(AppMsg,Sparse)
  else
    write(errstr,*) 'Last entry Y:  TIME, FRACEXP:',TIME(nr),FRACEXP(nr)
    call message(AppMsg,Sparse)
  endif

  call release(expotable)

  !-----------------------------------
  
  if (itest /= 1) then

    TimeCol  = column(eventstable, "TIME" ,read)
    FrameCol = column(eventstable, "FRAME",read)
    rawXCol  = column(eventstable, "RAWX" ,read)
    rawYCol  = column(eventstable, "RAWY" ,read)
    phaCol   = column(eventstable, "PHA"  ,read)
    flagCol  = column(eventstable, "FLAG" ,read)
    ccdnrCol = column(eventstable, "CCDNR",read)

    nevrows = numberofRows(eventstable)
    nr = nevrows-1

    rawX => int16Data(rawXCol)
    rawY => int16Data(rawYCol)
    pha  => int16Data(phaCol)
    flag => int32Data(flagCol)
    time => real64Data(timeCol)
    frame => int32Data(frameCol)
    ccdnr => int8Data(ccdnrCol)

    if (nevrows < 1) then
      write(errstr,*) 'Test not successful for ITEST = ',itest, &
         ', empty input file'
      error_in_test = .true. 
      call warning("emptyEventFile")
    elseif (nevrows .ne. 96832) then
      error_in_test = .true. 
      write(errstr,*) 'Event file does not have 96832 entries, but:',nevrows   
      call warning("wrongEventFileSize")
      call message(AppMsg,Sparse,&
           'Last entry: TIME, FRAME, RAWX, RAWY, PHA, FLAG, CCDNR')
      write(errstr,*) TIME(nr), FRAME(nr), RAWX(nr), &
           RAWY(nr), PHA(nr), FLAG(nr), CCDNR(nr)
      call message(AppMsg,Sparse)
      write(errstr,*) TIME(0), FRAME(0), RAWX(0), &
           RAWY(0), PHA(0), FLAG(0), CCDNR(0)
      call message(AppMsg,Sparse)
    else  
      call message(AppMsg,Sparse,&
           'Last entry: TIME, FRAME, RAWX, RAWY, PHA, FLAG, CCDNR')
      write(errstr,*) TIME(nr), FRAME(nr), RAWX(nr), &
           RAWY(nr), PHA(nr), FLAG(nr), CCDNR(nr)
      call message(AppMsg,Sparse)
    endif

    diff_ev_ex_Y = diff_ev_ex_Y-time(nr)

    write(errstr,*) "Time differences last entry: EXP-EVT (case N Y):",&
                     diff_ev_ex_N,diff_ev_ex_Y
    call message(AppMsg,Sparse)
 
    call message(AppMsg,Sparse,"Last event entries in N and Y cases:")
    write(errstr,*) TIME_N(nt+1), FRAME_N(nt+1), RAWX_N(nt+1), &
         RAWY_N(nt+1), PHA_N(nt+1), FLAG_N(nt+1), CCDNR_N(nt+1)
    call message(AppMsg,Sparse)
    write(errstr,*) TIME(nr), FRAME(nr), RAWX(nr), &
         RAWY(nr), PHA(nr), FLAG(nr), CCDNR(nr)
    call message(AppMsg,Sparse)

    call message(AppMsg,Sparse,'N Y : TIME FRAME RAWX RAWY PHA') 
    do i = 1,nt
      write(errstr,'(2f16.5,8i8)') &
          time_N(i),time(ns-1+i),frame_N(i),frame(ns-1+i), &
          rawx_N(i),rawx(ns-1+i),rawy_N(i),rawy(ns-1+i),pha_N(i),pha(ns-1+i)
      call message(AppMsg,Sparse)
    enddo

    do i = 0,  nevrows-1
      if (rawX(i) < 1 .or. rawY(i) < 1 .or. &
          rawX(i) > 64 .or. rawY(i) > 200) then
        if (flag(i) == 0) then
           write(errstr,*) "Coordinates out of range: rawX/Y(",i+1,")=(",&
                    rawX(i),",",rawY(i),") but not flagged: flag(",i,")=",flag(i)
           error_in_test = .true.
           call warning("wrongCoordsFlag")  
        endif 
      endif    
    enddo

    if (hasAttribute(eventstable,"EMPTYFRM")) then
      empty_frames = (trim(stringAttribute(eventstable,"EMPTYFRM")) == 'enabled')
    endif

    !---------------------------------------------
    write (*,*) 'testing ',gtilist(itest),' ...'
    !---------------------------------------------

    ingti = dataSet(gtilist(itest),read)

    !-----------------------------------

    if (nevrows > 0) then
      CCDno    =  ccdnr(0)
    else
      CCDno    =  6
    endif

    write(gtiname,'(a,i2.2)') "STDGTI",CCDno

    gtitable = table(ingti,gtiname)

    if (.not.hasColumn(gtitable,"START") .or. &
        .not.hasColumn(gtitable,"STOP")) &
        call error("noGTIColumn", &
                            "GTI file: column absent or incorrectly named")

    !-----------------------------------

    timeCol  = column(eventstable,"TIME",read)
    startCol = column(gtitable,  "START",read)
    stopCol  = column(gtitable,   "STOP",read)

    ngtirows = numberofRows(gtitable)

    if (ngtirows == 0) then 
      call warning("noGTIs","There are no GTI rows")
    else
      time  => real64Data(timeCol)
      start => real64Data(startCol)
      stop  => real64Data(stopCol)

      deltat = real64Attribute(gtitable,"FRMTIME")/1000.
      halfdt = deltat/2

      do i = 0, nevrows-1
         if (flag(i) == 0) then
!           if (abs(start(0) - (time(i) - halfdt)) > maxerr) &
            if (abs(start(0) - (time(i) - 2*halfdt)) > maxerr) then
              write(errstr,*) 'i, start(0), time(i), diff = ', &
                   i, start(0), time(i), &
                   abs(start(0) - (time(i) - 2*halfdt))
              call message(AppMsg,Sparse)
              call warning ("wrongGTI","Wrong GTI start calculation")
              if (.not. empty_frames) error_in_test = .true.
              exit
            endif
         endif
      enddo

      do i = nevrows-1, 0, -1
        if (flag(i) == 0) then
!           if (abs(stop(ngtirows-1) - (time(i) + halfdt)) > maxerr) &
            if (abs(stop(ngtirows-1) - (time(i) + 0*halfdt)) > maxerr) then
              write(errstr,*) 'i, stop(ngtirows-1), time(i), diff = ', &
                   i,stop(ngtirows-1) , time(i), &
                   abs(stop(ngtirows-1) - (time(i) + 0*halfdt))
              call message(AppMsg,Sparse)
              call warning ("wrongGTI","Wrong GTI stop calculation")
              if (.not. empty_frames) error_in_test = .true.
              exit
            endif
         endif
      enddo

    endif

    call release(ingti)

  endif  

  call release(inevents)

  if (error_in_test) then
    write(errstr,*) 'Test not successful for ITEST = ',itest
    call error("failure")
  else
    call message(AppMsg,Verbose,' ... testing successful')
    call error("artificialFailure", &
      "!! This error has been produced intentionally to obtain debug infos !!")
  endif

end
